---------------------------------------------
-- Control Unit for the 4-bit serial adder  
---------------------------------------------

LIBRARY IEEE;
   USE IEEE.STD_LOGIC_1164.ALL;
   USE IEEE.STD_LOGIC_ARITH.ALL;
   USE IEEE.STD_LOGIC_UNSIGNED.ALL;

ENTITY Control_Unit IS  
	PORT(
			CLK, CD 	:	IN  STD_LOGIC;	
			TC4			:	IN  STD_LOGIC;
			ST			:	IN  STD_LOGIC;
		    LDA			:   OUT STD_LOGIC_VECTOR(1 DOWNTO 0);
			LDW			:	OUT std_logic;
			LDS			:	OUT std_logic;
			Sel_Carry	:	OUT std_logic;
			EOP			:	OUT std_logic;
			CE  		:	OUT std_logic
         );
END Control_Unit;

ARCHITECTURE FSM_like OF Control_Unit IS

	TYPE State_type IS (Idle, Load_operands, Shift_accumulator, Shift_operands, Load_result, Set_EOP_Flag);  
--------> This is important: specifying to the synthesiser the code for the states
--	ATTRIBUTE syn_enum_encoding : STRING;
--	ATTRIBUTE syn_enum_encoding OF State_type : TYPE IS "sequential";
		-- (It may be:"sequential" (binary); "gray"; "one-hot", etc.
	SIGNAL present_state, future_state	: State_type ; 

		-- These two lines work OK, but the state machine is not recognised.
		-- and regular logic is synthesised (so it's not so good as the syn_encoding).  
-- ATTRIBUTE enum_encoding : string;
-- ATTRIBUTE enum_encoding of State_type : TYPE IS "sequential";


-- Constants for special states (the first and the last state)
CONSTANT Reset	: State_type := Idle; -- The fisrt state. This is another name for the state Num0

BEGIN
---------------------- State Register: the only clocked block.  
---------------------- The "memory" of the system (future events will depend on past events 

state_register: PROCESS (CD,CLK)
	BEGIN
						
	IF CD = '1' THEN 			-- asynchronous reset of the FSM 
		present_state <= Reset; 
	ELSIF (CLK'EVENT and CLK = '1') THEN	-- Synchronous register (D-type flip-flop)
		present_state <= future_state; 
	END IF;

END PROCESS state_register;


------------------------- CC1: Combinational system for calculating next state

CC_1: PROCESS (present_state, TC4, ST)
	BEGIN
		CASE present_state is
			when Idle =>
				if ST ='0' then 
					future_state <= Idle;
				else
					future_state <=	Load_operands ;
				end if;

			when Load_operands =>
					future_state <=	Shift_accumulator ;

			when Shift_accumulator => 
				 future_state <=Shift_operands;
				 
			when Shift_operands =>
			
				if TC4='1' then 
					future_state <= Load_Result;
				else 
					 future_state <= Shift_accumulator;			
				end if;

			when Load_result    => 
			         future_state <=Set_EOP_Flag;

			when Set_EOP_Flag	=>
				if ST ='1' then 
					future_state <= Set_EOP_Flag; -- The initial ST signal is not yet reset and, so, let's wait ... 
				else
					future_state <=	Idle ;
				end if;

		end case;

END PROCESS CC_1;


------------------------- CS_2: combinational system for calculating extra outputs 
-------------------------- and outputing the present state (the actual count)

CC_2: PROCESS (present_state)
	BEGIN

		CASE present_state is
			when Idle =>
				LDA 		<= "00";
				Sel_Carry 	<= '0';
				LDW 		<= '0';
				CE 			<= '0';
				LDS 		<= '0';
				EOP			<= '0';
				 
			when Load_operands =>
				LDA 		<= "11";
				Sel_Carry 	<= '0';
				LDW 		<= '0';
				CE 			<= '0';
				LDS 		<= '0';
				EOP			<= '0';  
				
			when Shift_accumulator =>
				LDA 		<= "00";
				Sel_Carry 	<= '1';
				LDW 		<= '1';
				CE 			<= '0';
				LDS 		<= '0';
				EOP			<= '0';

			when Shift_operands =>
				LDA 		<= "01";
				Sel_Carry 	<= '1';
				LDW 		<= '0';
				CE 			<= '1';
				LDS 		<= '0';
				EOP			<= '0';

			when Load_result => 
   
				LDA 		<= "00";
				Sel_Carry 	<= '0';
				LDW 		<= '0';
				CE 			<= '0';
				LDS 		<= '1';
				EOP			<= '0';

			when Set_EOP_Flag  => 

				LDA 		<= "00";
				Sel_Carry 	<= '0';
				LDW 		<= '0';
				CE 			<= '0';
				LDS 		<= '0';
				EOP			<= '1';


		end case;
	

END PROCESS CC_2;



-- Place other logic if necessary



END FSM_like;

